/*
  This is commands.h
  
  Coxeter version 1.0  Copyright (C) 1991 Fokko du Cloux
  See file main.c for full copyright notice
*/

#include "types.h"
#include "globals.h"

static weylarr_t input_array;

#define MAXMODE 5          /* number of modes */
#define MAXCELL 57         /* number of cells in command-tree */

#define MAXCOMMAND 16      /* number of commands */
#define MAXWEYLCMD 1       /* commands in weyl mode */
#define MAXBRCMD 1         /* commands in bruhat mode */
#define MAXKLCMD 4         /* commands in kl mode */

#define general_m 0        /* general mode */
#define help_m 1           /* help mode */
#define weyl_m 2           /* weyl mode */
#define bruhat_m 3         /* bruhat mode */
#define kl_m 4             /* kl mode */

#define undef_c -2         /* used for undefined commands */
#define ambig_c -1         /* used to detect ambiguous commands */
#define _c 0               /* carriage-return (empty command) */
#define author_c 1
#define cell_c 2
#define comp_c 3
#define help_c 4
#define info_c 5
#define kl_c 6
#define pol_c 7
#define pres_c 8
#define qq_c 9
#define quit_c 10
#define rank_c 11
#define rep_c 12
#define tex_c 13
#define type_c 14
#define verma_c 15

static int prev_mode;             /* remembers the previous mode */
static int prev_command;          /* remembers the previous command */

static int (*response[MAXCOMMAND])();

static ccell_t *comm_root;

int comm_tdata[]={10,
1/*a*/,2/*c*/,1/*h*/,1/*i*/,1/*k*/,2/*p*/,2/*q*/,2/*r*/,2/*t*/,1/*v*/,
1/*au*/,1/*ce*/,1/*co*/,1/*he*/,1/*in*/,0/*kl*/,1/*po*/,1/*pr*/,0/*qq*/,
  1/*qu*/,1/*ra*/,1/*re*/,1/*te*/,1/*ty*/,1/*ve*/,
1/*aut*/,1/*cel*/,1/*com*/,1/*hel*/,1/*inf*/,0/*pol*/,1/*pre*/,1/*qui*/,
  1/*ran*/,0/*rep*/,0/*tex*/,1/*typ*/,
  1/*ver*/,
1/*auth*/,0/*cell*/,0/*comp*/,0/*help*/,0/*info*/,0/*pres*/,0/*quit*/,
  0/*rank*/,0/*type*/,1/*verm*/,
1/*autho*/,0/*verma*/,
0/*author*/};

int comm_nbr[]={_c,
author_c/*a*/,ambig_c/*c*/,help_c/*h*/,info_c/*i*/,kl_c/*k*/,ambig_c/*p*/,
  quit_c/*q*/,ambig_c/*r*/,ambig_c/*t*/,verma_c/*v*/,
author_c/*au*/,cell_c/*ce*/,comp_c/*co*/,help_c/*he*/,info_c/*in*/,kl_c/*kl*/,
  pol_c/*po*/,pres_c/*pr*/,qq_c/*qq*/,quit_c/*qu*/,rank_c/*ra*/,rep_c/*re*/,
  tex_c/*te*/,type_c/*ty*/,verma_c/*ve*/,
author_c/*aut*/,cell_c/*cel*/,comp_c/*com*/,help_c/*hel*/,info_c/*inf*/,
  pol_c/*pol*/,pres_c/*pre*/,quit_c/*qui*/,rank_c/*ran*/,
  rep_c/*rep*/,tex_c/*tex*/,type_c/*typ*/,
  verma_c/*ver*/,
author_c/*auth*/,cell_c/*cell*/,comp_c/*comp*/,help_c/*help*/,info_c/*info*/,
  pres_c/*pres*/,quit_c/*quit*/,rank_c/*rank*/,type_c/*type*/,
  verma_c/*verm*/,
author_c/*autho*/,verma_c/*verma*/,
author_c/*author*/};

char comm_letter[]={'\000',
'a'/*a*/,'c'/*c*/,'h'/*h*/,'i'/*i*/,'k'/*k*/,'p'/*p*/,'q'/*q*/,'r'/*r*/,
  't'/*t*/,'v'/*v*/,
'u'/*au*/,'e'/*ce*/,'o'/*co*/,'e'/*he*/,'n'/*in*/,'l'/*kl*/,'o'/*po*/,
  'r'/*pr*/,'q'/*qq*/,'u'/*qu*/,
  'a'/*ra*/,'e'/*re*/,'e'/*te*/,'y'/*ty*/,'e'/*ve*/,
't'/*aut*/,'l'/*cel*/,'m'/*com*/,'l'/*hel*/,'f'/*inf*/,'l'/*pol*/,'e'/*pre*/,
  'i'/*qui*/,'n'/*ran*/,'p'/*rep*/,
  'x'/*tex*/,'p'/*typ*/,'r'/*ver*/,
'h'/*auth*/,'l'/*cell*/,'p'/*comp*/,'p'/*help*/,'o'/*info*/,'s'/*pres*/,
  't'/*quit*/,'k'/*rank*/,'e'/*type*/,'m'/*verm*/,
'o'/*autho*/,'a'/*verma*/,
'r'/*author*/};

/* Strings for the help facility */

char *mode_name[MAXMODE] = {"general","help","weyl","bruhat","kazhdan-lusztig"};

#define _tag "\n\tcarriage return : repeats previous command"
#define author_tag "\n\tauthor : prints authors' adress and other info"
#define cell_tag "\n\tcell : kl-pols in a given bruhat cell"
#define comp_tag "\n\tcomp : compute; finds normal forms"
#define help_tag "\n\thelp : you are using it now"
#define info_tag "\n\tinfo : prints general information about the program"
#define kl_tag "\n\tkl : computes the kl-files"
#define pol_tag "\n\tpol : kl-pol for two given elements in W"
#define pres_tag "\n\tpres : complete presentation"
#define qq_tag "\n\tqq : leaves the program"
#define quit_tag "\n\tquit : leaves present mode"
#define rank_tag "\n\trank : sets the rank"
#define rep_tag "\n\trep : representation; sets the input format"
#define tex_tag "\n\ttex : produces a TeX-file"
#define type_tag "\n\ttype : sets the type"
#define verma_tag "\n\tverma : multiplicities in Verma modules"

char *allcmds[MAXCOMMAND] = {_tag, author_tag, cell_tag, comp_tag, help_tag,
info_tag, kl_tag, pol_tag, pres_tag, qq_tag, quit_tag, rank_tag, 
rep_tag, type_tag, tex_tag, verma_tag};
char *weylcmds[MAXWEYLCMD] = {quit_tag};
char *bruhatcmds[MAXBRCMD] = {quit_tag};
char *klcmds[MAXKLCMD] = {quit_tag};

/* help messages */

char *_mess = "";
char *author_mess = "";

char *cell_mess = "\n\
This command will print all the Kazhdan-Lusztig polynomials for a given w.\n\
You can either output to a disk-file, in which case the output will be in\n\
TeX-format, or to the screen.\n\n";

char *comp_mess = "\n\
This command will \"compute\" its input; input to other commands is\n\
usually also piped through comp. You may enter group elements either\n\
as strings of generators, or by their number (in the lexicographical\n\
order). Such a number should start with a #. Two concatenated group\n\
elements, possibly separated by white space, are multiplied. Blanks,\n\
newlines and tabs are considered to be white space, and are ignored\n\
except inasmuch they act as separators (newlines must be escaped with\n\
\\, lest they terminate the input). The symbol ^ stands for \"power\";\n\
it must be followed by a number (which for simplicity we restrict to\n\
be strictly smaller than the order of the group). The symbol * stands\n\
for the longest element of the group. A group element immediately\n\
followed by * will be multiplied on the right by the longest element;\n\
i.e., in this case * is a unary operator following its argument. The\n\
symbol ! stands for \"inverse\"; it acts as a unary operator also. The\n\
three operators ^, * and ! have higher precedence than concatenation;\n\
they may be combined in any order.\n\
Clearly this does not cover al the desirable features, although any\n\
reasonable computation can be carried out in a few stages. The output of\n\
comp is the normal form of its input, followed by its number in\n\
lexicographical order (which can then be used as a shorthand to denote\n\
the element in question).\n\n";

char *help_mess = "";

char *help1_mess = "\nHere is a list of the available commands :\n";

char *help2_mess = "\n\n\
To obtain information about a command, enter its name. To exit the help\n\
facility, enter \"quit\" (another quit will make you exit the program).\n\
Note that command completion is enforced, as long as there is no ambiguity.\n\
To avoid confusion, only subsets of the full commands are allowed.\n\n";

char *info_mess = "\n\
The main purpose of this program is to give access to the Kazhdan-Lusztig\n\
polynomials of all the finite irreducible Coxeter groups of rank up to 6.\n\
We have also provided some additonal facilities for reducing and\n\
comparing group elements, and for finding multiplicities in Verma modules.\n\n\
The program should run fine on any workstation with a few megabytes of RAM\n\
(depending on the group under consideration; one megabyte is more than\n\
enough for F4); for the two hardest cases that we consider, namely H4 and\n\
E6, a few dozen megabytes of swap space will be needed. The program should\n\
run with only very minor modifications, if any, on a Macintosh (on which\n\
it was originally developed using Think C), although then, of course, swap\n\
space will not be available. To run it on a PC will be harder because of\n\
the weird adressing mode used by Intel processors\n\n";

char *kl_mess = "\n\
This command enters kl-mode; when it is invoked, the Kazhdan-Lusztig\n\
files are computed. These files contain in condensed form those\n\
kl-polynomials that cannot be deduced from others by shifting on the\n\
left or right by generators that leave the Bruhat-cell invariant\n\
(call the corresponding pairs \"extremal\"). In order not to recompute\n\
the files too often, and maybe also for future processing by other\n\
programs, they are dumped to the disk after computation. Subsequent\n\
calls to kl will just have to load them; unfortunately, this can\n\
still be quite a job because in the loading process a conversion\n\
must be made from relative to absolute adresses.\n\
For each group, three files are dumped, called \"store\", \"zone\",\n\
and \"list\" (not very great names). Zone contains the raw polynomials,\n\
written once each, in lexicographical order,  first the degree d,\n\
then d+1 coefficients; store contains a tree-structure, with root 1\n\
(the only polynomial of degree 0), reflecting the order in which\n\
the polynomials were found; list is a list of the extremal pairs,\n\
and the corresponding polynomials, and starts with the necessary\n\
data to read the files, e.g. the size of each of them.\n\
Clearly all this is very tentative and should be improved. For example\n\
loading could be made trivial if relative adresses were used throughout.\n\
The tree-structure should be optimized during the dumping-process. Also,\n\
store and zone together are very redundant.\n\n";

char *pol_mess = "\n\
This command prints a single Kazhdan-Lusztig polynomial P_{y,w}. The\n\
input is made of two group elements, y first, then w.\n\n";

char *pres_mess = "\n\
This command constructs a \"complete presentation\" of the group under\n\
consideration, i.e. a set of relations such that the normal form of any\n\
element is found by applying these relations left-to-right until no\n\
further reduction is possible. (It can be shown that these relations\n\
are unique, if we insist that the right-hand side should always be\n\
smaller in lexicographical order than the left-hand side, and if the\n\
presentation is minimal in the sense that no subexpression of any\n\
left-hand side can be reduced).\n\
Such a set of relations is exactly what is needed by a program such\n\
Mathematica; as a matter of fact, we have written our relations in\n\
Mathematica format, the generators being s[1], s[2], etc. However,\n\
practice is very disappointing : Mathematica turns out to be\n\
intolerably slow and inefficient, incapable of even loading these\n\
relations except in a few simple cases. Therefore the chief interest\n\
of these relations is theoretical.\n\n";

char *qq_mess = "\n\
Exits the program immediately.\n\n";

char *quit_mess = "\n\
Exits the current mode.\n\n";

char *rank_mess = "\n\
This command enables you to (re)define the rank of the Coxeter group you're\n\
interested in. For this, the type must have been set. We have allowed a few\n\
extensions to the usual restrictions on the rank, such as the groups of type\n\
D2, D3, E3, E4, E5; D2 and E3 are not irreducible (A1xA1 and A1xA2\n\
respectively) the others are already obtained as A3, A4, D5, but with a\n\
different numbering of the generators. The main purpose is to simplify the\n\
induction arguments.\n\n";

char *rep_mess = "\n\
(Re)defines the representation of generators for input-output\n\
purposes, i.e. digits, or lower-case letters.\n\n";

char *run_mess = "\n\
Starts the program.\n\n";

char *tex_mess = "\n\
Produces a TeX-file with the full output (i.e. all the Kazhdan-\n\
Lusztig polynomials for all the extremal pairs). Should be used\n\
with care, or it will overflow your disk (F4 produces some 270 pages\n\
of output; I wouldn't try H4 if I were you).\n\n";

char *type_mess = "\n\
This command enables you to (re)define the type of Coxeter group you're\n\
interested in. Currently, only finite irreducible Coxeter groups are\n\
allowed, and the dihedral non-cristallographic groups have been excluded\n\
for simplicity. We have followed the terminology from Bourbaki, so the\n\
infinite series are labelled A,B,D (C is the same as B), and the exceptional\n\
groups are of types E,F,G,H (the latter being the non-cristallographic\n\
case).\n\n";

char *verma_mess = "\n\
This command gives access to the multiplicities in Verma modules. We\n\
denote by M(w) (resp. L(w)) the Verma module (resp. the irreducible\n\
module) with highest weight w\\rho - \\rho -- notice that this is NOT the\n\
convention in Kazhdan and Lusztig's Inventiones paper. Then if w is given,\n\
the command prints the multiplicity (even if it is 0) for each y in the\n\
Weyl group. In order to save space, we don't print y; they just come in\n\
lexicographical order. Each line starts with the number of the first\n\
element for that line.\n\n";

/* Acknowledgements and adresses */

char *acknowledgement = "\
\n\tCoxeter version 1.0 Copyright (C) 1991 Fokko du Cloux\n\
\tEcole Polytechnique -- Centre de Mathematiques\n\
\t91128 Palaiseau Cedex, FRANCE\n\n\te-mail : ducloux@orphee.polytechnique.fr\n\n\
The Centre de Mathematiques is the Laboratoire Associe no. D-0169 du CNRS.\n\
It is a pleasure to acknowledge the hospitality and support of the\n\
University of Utah, where part of the work on this program was done during\n\
the winter of 1991.\n\n\
This program is made available under the terms of the GNU General Public\n\
License. In particular, it may be freely copied, modified and redistributed.\n\
Any program incorporating code taken from this program should make a\n\
statement to the same effect. More details and a full copy of the General\n\
Public License are located at the beginning of the source file main.c.\n\
If this is not available to you (as it should), a copy of the License \n\
can be obtained from the Free Software Foundation, Inc., 675 Mass Ave,\n\
Cambridge, MA 02139, USA.\n\n\
The author will be grateful for any comments and bug-reports concerning this\n\
program. However, there is no promise that any maintenance or further\n\
development will be undertaken. The sole purpose of this program is to serve\n\
the research interests of the author and those of other workers in the field.\n\n";


/* Functions local to command.c */

static Author();
static CarriageReturn();
static Cell();
static Compute();
static Dummy();
static FindCommand();
static GetCommand();
static HelpMe();
static Info();
static InitCommands();
static KLResponse();
static MakeCommandTree();
static MakeResponse();
static MakeSize();
static NewType();
static NewRank();
static NewRepresentation();
static Pol();
static Presentation();
static Prompt();
static QQuit();
static Quit();
static Tex();
static Start();
static VermaResponse();
static Unimplemented();
